from .. import *
from .base import *

class TText(TBase):
    def __init__(self, text, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.update(self.text)
        else:
            self.ui = widgets.Label(self.text, markup=False)
        super().update(prev)



from .. import *
from .base import *

class FRow(FBase):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ft.Row()
        self.ui.expand = self.layout_weight
        super().update(prev)

    def addChild(self, idx, child):
        self.ui.controls.insert(idx, child.outer)
        try:
            self.ui.update()
        except:
            pass

    def removeChild(self, idx, child):
        self.ui.controls.pop(idx)
        self.ui.update()

class FColumn(FBase):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ft.Column()
        self.ui.expand = self.layout_weight
        super().update(prev)

    def addChild(self, idx, child):
        self.ui.controls.insert(idx, child.outer)
        try:
            self.ui.update()
        except:
            pass

    def removeChild(self, idx, child):
        self.ui.controls.pop(idx)
        self.ui.update()

class FSpacer(FBase):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ft.Container()
        super().update(prev)



from .. import *
from .base import *

class FApplication(PUIView):
    def __init__(self):
        super().__init__()
        self.ready = False

    def update(self, prev=None):
        if not self.ready:
            return
        super().update(prev)

    def flet_app(self, page: ft.Page):
        self.ui = page
        self.ready = True
        self.update()

    def addChild(self, idx, child):
        if idx > 0:
            print("Flet backend only support single window")

    def removeChild(self, idx, child):
        pass

    def start(self):
        ft.app(self.flet_app)



from .. import *
from .base import *
class TkButton(TkBaseWidget):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.config(text = self.text)
            self.ui.config(command = self._clicked)
        else:
            self.ui = tk.Button(self.tkparent.inner, text=self.text, command=self._clicked)
        super().update(prev)



# dprint = print
dprint = lambda *x: x

def recur_delete(node, child, direct):
    child.destroyed = True
    for sc in child.children:
        recur_delete(child, sc, False)
    child.destroy(direct)

def sync(node, oldDOM, newDOM):
    dprint("syncing", node.key, len(oldDOM), len(newDOM))

    dprint("  ===OLD===")
    for c in oldDOM:
        dprint("   ", c.key)

    dprint("  ===NEW===")
    for c in newDOM:
        dprint("   ", c.key)
    oldMap = [x.key for x in oldDOM]
    newMap = [x.key for x in newDOM]

    tbd = []
    for i,new in enumerate(newDOM):
        while True:
            if i < len(oldDOM) and oldMap[i] == new.key: # matched
                dprint(f"MATCHED {i} {new.key}")
                old = oldDOM[i]

                try:
                    new.update(old)
                except:
                    import traceback
                    print("## <ERROR OF update() >")
                    print(node.key)
                    traceback.print_exc()
                    print("## </ERROR OF update()>")

                if not new.terminal:
                    sync(new, old.children, new.children)

                break # finish

            trimmed = False
            while i < len(oldDOM) and not oldDOM[i].key in newMap: # trim old nodes
                dprint(f"TRIM {i} {oldDOM[i].key}")
                old = oldDOM.pop(i)
                oldMap.pop(i)
                node.removeChild(i, old)
                tbd.append(old)
                trimmed = True

            if trimmed:
                continue # restart

            try: # node exists
                idx = oldMap[i+1:].index(new.key)+i+1
                if idx==i+1: # move wrong node
                    oldMap.pop(i)
                    old = oldDOM.pop(i)
                    node.removeChild(i, old)

                    node.addChild(len(oldDOM), old)
                    oldMap.append(old.key)
                    oldDOM.append(old)

                    continue # restart
                else: # move target node
                    oldMap.pop(idx)
                    old = oldDOM.pop(idx)
                    node.removeChild(idx, old)

                    try:
                        new.update(old)
                    except:
                        import traceback
                        print("## <ERROR OF update() >")
                        print(node.key)
                        traceback.print_exc()
                        print("## </ERROR OF update()>")

                    if not new.terminal:
                        sync(new, old.children, new.children)

                    node.addChild(i, new)
                    oldDOM.insert(i, None) # placeholder
                    oldMap.insert(i, new.key)

            except: # new node
                try:
                    new.update(None)
                except:
                    import traceback
                    print("## <ERROR OF update() >")
                    print(new.key)
                    traceback.print_exc()
                    print("## </ERROR OF update()>")
                if not new.terminal:
                    sync(new, [], new.children)
                node.addChild(i, new)
                oldDOM.insert(i, None) # placeholder
                oldMap.insert(i, new.key)

            break # finish

    nl = len(newDOM)
    while len(oldDOM) > nl:
        old = oldDOM.pop(nl)
        oldMap.pop(nl)
        node.removeChild(nl, old)
        tbd.append(old)

    for c in newDOM:
        c.postUpdate()

    node.postSync()

    for old in tbd:
        recur_delete(node, old, True)



from .. import *
from .base import *
import functools

class TkApplication(PUIView):
    def redraw(self):
        if self.ui:
            self.ui.after(0, functools.partial(self.update))
        else:
            self.update()

    def update(self, prev=None):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = tk.Tk()
            ttkStyle = ttk.Style(self.ui)
            ttkStyle.theme_use('classic') # macOS's aqua doesn't respect background setting for some widgets
            self.ui.withdraw()

        super().update(prev)

    def addChild(self, idx, child):
        pass

    def removeChild(self, idx, child):
        pass

    def start(self):
        self.ui.mainloop()



from .. import *
from .base import *

class QtHBox(QtBaseLayout):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QHBoxLayout()
        super().update(prev)

class QtVBox(QtBaseLayout):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QVBoxLayout()
        super().update(prev)

class QtSpacerItem(PUINode):
    terminal = True

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        super().update(prev)

    def destroy(self, direct):
        # self.ui.deleteLater() # QSpacerItem doesn't have .deleteLater()
        self.ui = None
        super().destroy(direct)



from .. import *
from .base import *

class FTextField(FBase):
    def __init__(self, model, label="", **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.label = label

    def update(self, prev):
        value = self.model.value
        if prev and prev.ui:
            self.ui = prev.ui
            if prev.last_value != value:
                self.ui.value = str(value)
                try:
                    self.ui.update()
                except:
                    pass
            self.last_value = value
        else:
            self.last_value = value
            self.ui = ft.TextField(label=self.label, value=value, on_change=self.on_textbox_changed, expand=self.layout_weight)
        super().update(prev)

    def on_textbox_changed(self, e):
        self.model.value = e.control.value



from .. import *
from .base import *

class TkWindow(TkBaseWidget):
    def __init__(self, title=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.title = title
        self.size = size
        self.curr_size = None
        self.maximize = maximize
        self.curr_maximize = None
        self.fullscreen = fullscreen
        self.curr_fullscreen = None

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.curr_size = prev.curr_size
            self.curr_maximize = prev.curr_maximize
            self.curr_fullscreen = prev.curr_fullscreen
        else:
            self.ui = tk.Toplevel(self.parent.inner)

        if self.curr_size != self.size:
            self.curr_size = self.size
            self.ui.geometry("x".join([str(v) for v in self.size]))
        if self.curr_maximize !=  self.maximize:
            self.curr_maximize =  self.maximize

            # https://stackoverflow.com/a/70061942
            try:
                self.ui.attributes('-zoomed', True)
            except:
                self.ui.state("zoomed")

        if self.curr_fullscreen != self.fullscreen:
            self.curr_fullscreen = self.fullscreen
            self.ui.attributes('-fullscreen', True)
        if not self.title is None:
            self.ui.title(self.title)

    def addChild(self, idx, child):
        if idx:
            return
        child.ui.pack(fill=tk.BOTH, expand=True)

    def removeChild(self, idx, child):
        child.ui.pack_forget()



import threading
from .utils import *

tls = threading.local()

class PuiViewNotFoundError(Exception): pass

def find_puiview():
    try:
        return tls.puistack[-1]
    except:
        raise PuiViewNotFoundError()

class PUINode():
    supported = True
    terminal = False
    def __init__(self, *args):
        from .view import PUIView

        if not hasattr(self, "name"):
            self.name = None

        self.destroyed = False
        self.retired_by = None
        self._debug = 0
        self._tag = ""

        self.layout_weight = None
        self.layout_width = None
        self.layout_height = None
        self.layout_padding = None
        self.layout_margin = None
        self.style_color = None
        self.style_bgcolor = None
        self.style_fontsize = None
        self.style_fontweight = None
        self.style_fontfamily = None

        self.onClicked = None

        self.ui = None
        self.args = args
        try:
            self.root = find_puiview()
            self.parent = self.root.frames[-1]
        except PuiViewNotFoundError:
            self.root = self
            self.parent = self

        if isinstance(self, PUIView):
            self.root = self

        self.genKey()

        self.children = []

        if self.parent is self:
            self.path = tuple()
        else:
            self.path = self.parent.path + tuple([len(self.parent.children)])
            self.parent.children.append(self)

        # print(type(self).__name__, self.path, "parent=", self.parent.path)

    def genKey(self):
        # key has to be relative to PUIView, so that it can be identical when a sub-PUIView is updated individually
        self.key = "|".join([x.name or type(x).__name__ for x in self.root.frames]+[self.name or type(self).__name__]+[str(id(x)) for x in self.args])
        if self._tag:
            self.key += f"@{self._tag}"

    def __enter__(self):
        # print("enter", type(self).__name__, id(self))
        self.root.frames.append(self)
        return self

    def __exit__(self, ex_type, value, traceback):
        # print("exit", type(self).__name__, id(self))
        self.root.frames.pop()
        if ex_type is None: # don't consume exception
            return self

    @property
    def inner(self):
        if self.ui:
            return self.ui
        return self.parent.inner

    @property
    def outer(self):
        if self.ui:
            return self.ui
        if self.children:
            return self.children[0].outer
        return None

    def comment(self):
        return None

    def update(self, prev):
        if prev:
            prev.retired_by = self

    def postUpdate(self):
        pass

    def postSync(self):
        pass

    def destroy(self, direct):
        self.root = None
        self.parent = None

    def addChild(self, idx, child):
        pass

    def removeChild(self, idx, child):
        pass

    def debug(self, level=1):
        self._debug = level
        return self

    def tag(self, name):
        self._tag = name
        self.genKey()
        return self

    def get_node(self):
        node = self
        while node.retired_by:
            node = node.retired_by
        return node

    def __repr__(self):
        segs = []
        headline = [
            "  "*len(self.path),
            self.name or type(self).__name__,
            # f"@{str(id(self))}", # print view id
            " {",
        ]

        # print view key
        headline.append(" # ")
        headline.append(self.key)

        headline.append("\n")
        segs.append("".join(headline))

        comment = self.comment()
        if comment:
            segs.append("  "*(len(self.path)+1))
            segs.append("# ")
            segs.append(comment)
            segs.append("\n")

        for i,c in enumerate(self.children):
            if i > 0:
                segs.append(",\n")
            segs.append(c.__repr__())
        segs.append("\n")
        segs.append("".join(["  "*len(self.path), "}"]))
        return "".join(segs)

    def layout(self, width=None, height=None, weight=None, padding=None, margin=None):
        if not width is None:
            self.layout_width = width
        if not height is None:
            self.layout_height = height
        if not weight is None:
            self.layout_weight = weight
        if not padding is None:
            self.layout_padding = trbl(padding)
        if not margin is None:
            self.layout_margin = trbl(margin)

        return self

    def style(self, color=None, bgColor=None, fontSize=None, fontWeight=None, fontFamily=None):
        if not color is None:
            self.style_color = color
        if not bgColor is None:
            self.style_bgcolor = bgColor
        if not fontSize is None:
            self.style_fontsize = fontSize
        if not fontWeight is None:
            self.style_fontweight = fontWeight
        if not fontFamily is None:
            self.style_fontfamily = fontFamily

        return self

    def click(self, callback, *cb_args, **cb_kwargs):
        self.onClicked = callback
        self.click_args = cb_args
        self.click_kwargs = cb_kwargs
        return self

    def _clicked(self, *args, **kwargs):
        node = self.get_node()
        if node.onClicked:
            node.onClicked(*self.click_args, **self.click_kwargs)

    def flet(self, **kwargs):
        return self

    def textual(self, **kwargs):
        return self

    def tk(self, **kwargs):
        return self

    def qt(self, **kwargs):
        return self



from typing import Type
from textual.driver import Driver
from .. import *
from .base import *
from textual.app import App, CSSPathType, ComposeResult
from textual.containers import Vertical
from textual.widgets import Button, Checkbox, Input, RadioButton

class PUIApp(App):

    def __init__(self, puiview, driver_class: Type[Driver] | None = None, css_path: CSSPathType | None = None, watch_css: bool = False):
        super().__init__(driver_class, css_path, watch_css)
        self.puiview = puiview

    def on_mount(self) -> None:
        self.puiview.redraw()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        event.button.puinode._clicked()

    def on_radio_button_changed(self, event: RadioButton.Changed) -> None:
        event.radio_button.puinode._changed(event.value)

    def on_input_changed(self, event: Input.Changed) -> None:
        event.input.puinode._changed(event.value)

    def on_input_submitted(self, event: Input.Submitted) -> None:
        event.input.puinode._submitted(event.value)

    def on_checkbox_changed(self, event: Checkbox.Changed) -> None:
        event.checkbox.puinode._changed(event.value)

    def compose(self) -> ComposeResult:
        yield Vertical(id="frame")

class TApplication(PUIView):
    def __init__(self):
        super().__init__()
        self.ui = PUIApp(self)

    def addChild(self, idx, child):
        if idx>0:
            raise RuntimeError("Textual backend only support single window")
        self.ui.query_one("#frame").mount(child.outer)

    def redraw(self):
        self.dirty = True
        if self.updating:
            return
        self.updating = True
        self.ui.call_next(self._redraw)

    def _redraw(self):
        with self.ui.batch_update():
            self.update()
        self.updating = False

    def run(self):
        # self.redraw() # need to be after on_mount
        self.start()

    def start(self):
        self.ui.run()



from .. import *
from .base import *
class TkLabel(TkBaseWidget):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.config(text = self.text)
        else:
            self.ui = tk.Label(self.tkparent.inner, text=self.text, anchor="w", justify="left")
            self.ui.bind("<Button-1>", self._clicked)
        if self.onClicked:
            self.ui.config(cursor="")
        super().update(prev)


from .. import *
from .base import *

class TRadioButton(TBase):
    def __init__(self, text, value, model):
        super().__init__()
        self.text = text
        self.value = value
        self.model = model

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = widgets.RadioButton(self.text)
        self.ui.puinode = self
        self.ui.value = self.value  == self.model.value
        super().update(prev)

    def _changed(self, value):
        if value:
            self.model.value = self.value
        else:
            self.root.redraw()


from .view import *
from collections import defaultdict

class DummyBinding():
    def __init__(self, value):
        self.value = value

class AttrBinding():
    def __init__(self, state, key):
        try:
            self.viewroot = find_puiview()
            self.viewparent = self.viewroot.frames[-1]
        except:
            pass
        self.state = state
        self.key = key
        dt = type(getattr(self.state, self.key))
        if dt is str:
            self.func = str
        elif dt is int:
            self.func = int
        elif dt is float:
            self.func = float
        else:
            self.func = lambda x:x

    @property
    def value(self):
        return getattr(self.state, self.key)

    @value.setter
    def value(self, value):
        try: # skip validation error
            setattr(self.state, self.key, self.func(value))
        except:
            pass

    def change(self, callback):
        getattr(self.state, "_StateObject__callbacks")[self.key].add(callback)

    def bind(self, getter, setter):
        getattr(self.state, "_StateObject__binders")[self.key] = None
        setattr(getattr(self.state, "_StateObject__values"), self.key, getter())
        getattr(self.state, "_StateObject__binders")[self.key] = (getter, setter)

class ListBinding():
    def __init__(self, state, key):
        try:
            self.viewroot = find_puiview()
            self.viewparent = self.viewroot.frames[-1]
        except PuiViewNotFoundError:
            pass
        self.state = state
        self.key = key
        dt = type(self.state[self.key])
        if dt is str:
            self.func = str
        elif dt is int:
            self.func = int
        elif dt is float:
            self.func = float
        else:
            self.func = lambda x:x

    @property
    def value(self):
        try: # skip validation error
            return self.state[self.key]
        except:
            pass

    @value.setter
    def value(self, value):
        self.state[self.key] = self.func(value)

    def change(self, callback):
        getattr(self.state, "_StateList__callbacks")[self.key].add(callback)

    def bind(self, getter, setter):
        getattr(self.state, "_StateList__binders")[self.key] = None
        self.state[self.key] = getter()
        getattr(self.state, "_StateList__binders")[self.key] = (getter, setter)

class DictBinding():
    def __init__(self, state, key):
        try:
            self.viewroot = find_puiview()
            self.viewparent = self.viewroot.frames[-1]
        except PuiViewNotFoundError:
            pass
        self.state = state
        self.key = key
        dt = type(self.state[self.key])
        if dt is str:
            self.func = str
        elif dt is int:
            self.func = int
        elif dt is float:
            self.func = float
        else:
            self.func = lambda x:x

    @property
    def value(self):
        return self.state[self.key]

    @value.setter
    def value(self, value):
        try: # skip validation error
            self.state[self.key] = self.func(value)
        except:
            pass

    def change(self, callback):
        getattr(self.state, "_StateDict__callbacks")[self.key].add(callback)

    def bind(self, getter, setter):
        getattr(self.state, "_StateDict__binders")[self.key] = None
        self.state[self.key] = getter()
        getattr(self.state, "_StateDict__binders")[self.key] = (getter, setter)

def _notify(listeners):
    tbd = []
    for l in listeners:
        if l.retired_by:
            tbd.append(l)
    for l in tbd:
        listeners.remove(l)
    for l in listeners:
        l.redraw()

class BaseState():
    pass

def State(data=None):
    if data is None:
        return StateObject()
    if isinstance(data, list):
        return StateList(data)
    if isinstance(data, dict):
        return StateDict(data)
    return StateObject(data)

class StateObject(BaseState):
    def __init__(self, values=None):
        self.__listeners = set()
        self.__callbacks = defaultdict(set)
        self.__binders = {}
        if values is None:
            self.__values = BaseState()
        else:
            self.__values = values

    def __call__(self, key=None):
        return AttrBinding(self, key)

    def __getattr__(self, key):
        if not key.startswith("_"):
            try:
                view = find_puiview()
                self.__listeners.add(view)
            except PuiViewNotFoundError:
                pass
        return getattr(self.__values, key)

    def __setattr__(self, key, value):
        if key.startswith("_"):
            object.__setattr__(self, key, value)
        else:
            if self.__binders.get(key):
                self.__binders[key][1](value)

            if type(value) is list:
                value = StateList(value)
            elif type(value) is dict:
                value = StateDict(value)
            if not hasattr(self.__values, key) or getattr(self.__values, key) != value:
                setattr(self.__values, key, value)
                _notify(self.__listeners)
                for cb in self.__callbacks[key]:
                    cb(value)

    def __repr__(self):
        return f"StateObject({self.__values.__repr__()})"

class StateList(BaseState):
    def __init__(self, values=None):
        self.__listeners = set()
        self.__callbacks = defaultdict(set)
        self.__binders = {}
        if values is None:
            self.__values = []
        else:
            self.__values = values

    def __call__(self, key=None):
        return ListBinding(self, key)

    def __getitem__(self, key):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass

        return self.__values[key]

    def __setitem__(self, key, value):
        if self.__binders.get(key):
            self.__binders[key][1](value)
        new = key >= len(self.__values)
        if new or self.__values[key] != value:
            self.__values[key] = value
            _notify(self.__listeners)
            for cb in self.__callbacks[key]:
                cb(value)
            if new:
                for cb in self.__callbacks[None]:
                    cb(value)

    def __bool__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return bool(self.__values)

    def __len__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return len(self.__values)

    def __iter__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__iter__()

    def __repr__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__repr__()

    def append(self, obj):
        self.__values.append(obj)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def clear(self):
        self.__values.clear()
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def count(self, value):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.count(value)

    def extend(self, iterable):
        self.__values.extend(iterable)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def index(self, value, *args, **kwargs):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.index(value, *args, **kwargs)

    def insert(self, index, object):
        self.__values.insert(index, object)
        _notify(self.__listeners)
        for cb in self.__callbacks[index]:
            cb(object)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def pop(self, index=-1):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        r = self.__values.pop(index)
        _notify(self.__listeners)
        for cb in self.__callbacks[index]:
            cb(self.__values)
        for cb in self.__callbacks[None]:
            cb(self.__values)
        return r

    def remove(self, value):
        self.__values.remove(value)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def reverse(self, value):
        self.__values.reverse(value)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)


    def sort(self, *args, **kwargs):
        self.__values.sort(*args, **kwargs)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def get(self, index, default=None):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        if index >= 0 and index < len(self.__values):
            return self.__values[index]
        else:
            return default

    def range(self):
        return range(len(self.__values))

class StateDict(BaseState):
    def __init__(self, values=None):
        self.__listeners = set()
        self.__callbacks = defaultdict(set)
        self.__binders = {}
        if values is None:
            self.__values = {}
        else:
            self.__values = values

    def __call__(self, key=None):
        return DictBinding(self, key)

    def __delitem__(self, key):
        self.__values.__delitem__(key)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def __getitem__(self, key):
        try:
            view = find_puiview()
            self.__listeners[key].add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values[key]

    def __getattr__(self, key):
        if not key.startswith("_"):
            try:
                view = find_puiview()
                self.__listeners[key].add(view)
            except PuiViewNotFoundError:
                pass
        return getattr(self.__values, key)


    def __setattr__(self, key, value):
        if key.startswith("_"):
            object.__setattr__(self, key, value)
        else:
            if self.__binders.get(key):
                self.__binders[key][1](value)
            _notify(self.__listeners)
            return setattr(self.__values, key, value)

    def __bool__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return bool(self.__values)

    def __iter__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__iter__()

    def __repr__(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.__repr__()

    def __setitem__(self, key, value):
        if self.__binders.get(key):
            self.__binders[key][1](value)
        if not key in self.__values or self.__values[key] != value:
            self.__values[key] = value
            _notify(self.__listeners)
        for cb in self.__callbacks[key]:
            cb(value)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def clear(self):
        self.__values.clear()
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)

    def get(self, key, default=None):
        try:
            view = find_puiview()
            self.__listeners[key].add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.get(key, default)

    def items(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.items()

    def keys(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.keys()

    def pop(self, key, default=None):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        r = self.__values.pop(key, default)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)
        return r

    def setdefault(self, key, default=None):
        r = self.__values.setdefault(key, default)
        _notify(self.__listeners)
        for cb in self.__callbacks[None]:
            cb(self.__values)
        return r

    def values(self):
        try:
            view = find_puiview()
            self.__listeners.add(view)
        except PuiViewNotFoundError:
            pass
        return self.__values.values()



from .. import *
from .base import *

class QtComboBox(QtBaseWidget):
    terminal = False
    def __init__(self, editable=False, index_model=None, text_model=None):
        super().__init__()
        self.editable = editable
        if index_model is None:
            index_model = DummyBinding(0)
        if text_model is None:
            text_model = DummyBinding("")
        self.index_model = index_model
        self.text_model = text_model
        self.index_changed = None
        self.text_changed = None

    def on(self, index_changed=None, text_changed=None):
        if index_changed:
            self.index_changed = index_changed
        if text_changed:
            self.text_changed = text_changed
        return self

    def update(self, prev):
        index = self.index_model.value
        text = self.text_model.value
        if prev and prev.ui:
            self.ui = prev.ui
            if prev.last_index != index:
                self.ui.currentIndexChanged.disconnect()
                self.ui.setCurrentIndex(index)
                self.ui.currentIndexChanged.connect(self.on_currentIndexChanged)
            self.last_index = index

            if prev.last_text != text:
                self.ui.currentTextChanged.disconnect()
                self.ui.setCurrentText(text)
                self.ui.currentTextChanged.connect(self.on_currentTextChanged)
            self.last_text = text
        else:
            self.last_index = index
            self.last_text = text
            self.ui = QtWidgets.QComboBox()
            self.ui.currentIndexChanged.connect(self.on_currentIndexChanged)
            self.ui.currentTextChanged.connect(self.on_currentTextChanged)
        self.ui.setEditable(self.editable)
        super().update(prev)

    def on_currentIndexChanged(self, idx):
        self.index_model.value = idx
        if self.index_changed:
            self.index_changed(idx)

    def on_currentTextChanged(self, text):
        self.text_model.value = text
        if self.text_changed:
            self.text_changed(text)

    def addChild(self, idx, child):
        self.ui.insertItem(idx, child.text)

    def removeChild(self, idx, child):
        self.ui.removeItem(idx)

class QtComboBoxItem(PUINode):
    def __init__(self, text):
        super().__init__()
        self.text = text



from .. import *
from .base import *

class QtButton(QtBaseWidget):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.setText(self.text)
            try:
                self.ui.clicked.disconnect()
            except:
                pass
            prev.callback = None
        else:
            self.ui = QtWidgets.QPushButton(text=self.text)
        self.ui.clicked.connect(self._clicked)
        super().update(prev)



from .. import *
from .base import *

class FCheckbox(FBase):
    def __init__(self, text, model):
        super().__init__()
        self.text = text
        self.model = model

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.value = self.model.value
            self.ui.on_change = self._changed
            self.ui.update()
        else:
            self.ui = ft.Checkbox(label=self.text, value=self.model.value, on_change=self._changed)

        super().update(prev)

    def _changed(self, event):
        node = self.get_node()
        self.model.value = node.ui.value



from .application import *
from .button import *
from .canvas import *
from .checkbox import *
from .label import *
from .layout import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .text import *
from .textfield import *
from .window import *
from .. import NotImplementedNode

Application = FApplication
Window = FWindow
HBox = FRow
VBox = FColumn
Label = FLabel
Button = FElevatedButton
Checkbox = FCheckbox
RadioButton = FRadio
Canvas = FCanvas
TextField = FTextField
ProgressBar = FProgressBar
Scroll = FScroll
Spacer = FSpacer
Text = FText
Html = FHtml
MarkDown = FMarkDown
Combobox = NotImplementedNode
ComboboxItem = NotImplementedNode

PUI_BACKEND = "flet"



from .. import *
from .base import *

class FLabel(FBase):
    def __init__(self, text, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ft.Text(spans=[])
        self.ui.expand = self.layout_weight
        if self.onClicked:
            self.ui.spans = [
                ft.TextSpan(self.text, on_click=self._clicked)
            ]
        else:
            self.ui.spans = [
                ft.TextSpan(self.text)
            ]
        try:
            self.ui.update()
        except:
            pass
        super().update(prev)


from .. import *
from textual import widgets, containers

class TBase(PUINode):
    container_x = False
    container_y = False
    strong_expand_x = False
    weak_expand_x = False
    strong_expand_y = False
    weak_expand_y = False
    weak_hug_x = False
    weak_hug_y = False
    expanding_x_children = 0
    expanding_y_children = 0

    @property
    def expand_x(self):
        return self.strong_expand_x or (self.weak_expand_x and not self.weak_hug_x)

    @property
    def expand_y(self):
        return self.strong_expand_y or (self.weak_expand_y and not self.weak_hug_y)

    def tremove(self):
        self.ui.remove()

    def destroy(self, direct):
        self.ui.remove()
        return super().destroy(direct)

    def update(self, prev):
        parent = self.tparent
        if parent:
            # request expanding from inside
            if parent.container_x:
                if self.layout_weight:
                    self.strong_expand_x = True
                    parent.expanding_x_children += 1
                    p = parent
                    while p:
                        if isinstance(p, TBase):
                            p.weak_expand_x = True
                        if p==p.parent:
                            break
                        p = p.parent
            if parent.container_y:
                if self.layout_weight:
                    self.strong_expand_y = True
                    parent.expanding_y_children += 1
                    p = parent
                    while p:
                        if isinstance(p, TBase):
                            p.weak_expand_y = True
                        if p==p.parent:
                            break
                        p = p.parent

            if parent.expanding_x_children > 0:
                self.weak_hug_x = True
            if parent.expanding_y_children > 0:
                self.weak_hug_y = True

        super().update(prev)

    def postUpdate(self):
        super().postUpdate()


        self.t_update_layout()

    @property
    def tparent(self):
        parent = self.parent
        while not isinstance(parent, TBase):
            if parent==parent.parent:
                parent = None
                break
            parent = parent.parent
        return parent

    def t_update_layout(self):
        if not self.ui:
            return

        width = "auto"
        if self.expand_x:
            width = "1fr"
        width = f"width:{width};"

        height = "auto"
        if self.expand_y:
            height = "1fr"
        height = f"height:{height};"

        if self._debug:
            print("layout", self.key, f"expand_x={self.expand_x}", f"expand_y={self.expand_y}", f"strong_x={self.strong_expand_x}", f"weak_x={self.weak_expand_x}", f"hug_x={self.weak_hug_x}", f"strong_y={self.strong_expand_y}", f"weak_y={self.weak_expand_y}", f"hug_y={self.weak_hug_y}")
        self.ui.set_styles(width+height)



from .. import *
from .base import *

class TkEntry(TkBaseWidget):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def update(self, prev):
        value = self.model.value
        if prev and prev.ui:
            self.variable = prev.variable
            self.ui = prev.ui
            if prev.last_value != value:
                self.variable.set(value)
            self.last_value = value
        else:
            self.variable = tk.StringVar(self.tkparent.inner, str(value))
            self.variable.trace_add("write", self.on_variable_changed)
            self.last_value = value
            self.ui = tk.Entry(self.tkparent.inner, textvariable=self.variable)
        super().update(prev)

    def on_variable_changed(self, var, index, mode):
        value = self.variable.get()
        self.model.value = value



from .. import *
from .base import *

class QtSplitter(QtBaseWidget):
    terminal = False
    def __init__(self, vertical=False):
        super().__init__()
        self.vertical = vertical

    def destroy(self, direct):
        if direct:
            for frame in self.frame:
                if frame:
                    frame.deleteLater()
        super().destroy(direct)

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.frame = prev.frame
        else:
            self.ui = QtWidgets.QSplitter()
            self.frame = []
        self.ui.setOrientation(QtCore.Qt.Orientation.Vertical if self.vertical else QtCore.Qt.Orientation.Horizontal)
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, QtBaseLayout):
            frame = QtWidgets.QWidget()
            frame.setLayout(child.outer)
            self.frame.insert(idx, frame)
            self.ui.insertWidget(idx, frame)
        elif isinstance(child, QtBaseWidget):
            self.ui.insertWidget(idx, child.outer)
            self.frame.insert(idx, None)
        else:
            self.addChild(idx, child.children[0])
    
    def removeChild(self, idx, child):
        child.outer.setParent(None)
        frame = self.frame.pop(idx)
        if frame:
            frame.deleteLater()



from .. import *
from .base import *
from tkinter.scrolledtext import ScrolledText

class ScrollableFrame(ttk.Frame):
    def __init__(self, parent):
        self.vertical = None
        self.horizontal = False

        self.container = ttk.Frame(parent)
        self.container.rowconfigure(0, weight=1)
        self.container.columnconfigure(0, weight=1)

        self.scroll_y = tk.Scrollbar(self.container, orient=tk.VERTICAL)
        self.scroll_x = tk.Scrollbar(self.container, orient=tk.HORIZONTAL)
        self.should_scroll_y = False
        self.should_scroll_x = False

        self.canvas = tk.Canvas(self.container)
        self.canvas.grid(row=0, column=0, sticky="nsew")
        super().__init__(self.canvas)
        self.canvas.create_window(0, 0, window=self, anchor="nw")

        self.bind("<Configure>", self._on_configure)

    def set_scroll(self, vertical, horizontal):
        self.vertical = vertical
        self.horizontal = horizontal

        self.update_scroller()

    def grid(self, *args, **kwargs):
        self.container.grid(*args, **kwargs)

    def configure(self, *args, **kwargs):
        self.container.configure(*args, **kwargs)

    def _on_configure(self, event):
        self.update_scroller()

    def update_scroller(self):
        # self.winfo_toplevel().update_idletasks()

        bbox = self.canvas.bbox("all")
        _, _, w, h = bbox
        self.canvas.configure(scrollregion=bbox)
        if self.horizontal is False:
            self.canvas.configure(width=w)
        if self.vertical is False:
            self.canvas.configure(height=h)

        self.should_scroll_y = self.vertical or (self.vertical is None and h > self.canvas.winfo_height())
        if self.should_scroll_y:
            self.scroll_y.config(command=self.canvas.yview)
            self.canvas.config(yscrollcommand=self.scroll_y.set)
            self.scroll_y.grid(row=0, column=1, sticky="ns")

            self.canvas.bind_all("<MouseWheel>", self._on_mousewheel_y)
            self.canvas.bind("<Button-4>", self._on_mousewheel_y)
            self.canvas.bind("<Button-5>", self._on_mousewheel_y)
        else:
            self.scroll_y.grid_forget()

        self.should_scroll_x = self.horizontal or (self.horizontal is None and w > self.canvas.winfo_width())
        if self.should_scroll_x:
            self.scroll_x.config(command=self.canvas.xview)
            self.canvas.config(xscrollcommand=self.scroll_x.set)
            self.scroll_x.grid(row=1, column=0, sticky="ew")
        else:
            self.scroll_x.grid_forget()

    def _on_mousewheel_y(self, event):
        if not self.should_scroll_y:
            return
        if event.num == 5:
            delta = 1
        elif event.num == 4:
            delta = -1
        elif event.delta > 0:
            delta = 1
        elif event.delta < 0:
            delta = -1
        self.canvas.yview_scroll(delta, "units")

class TkScroll(TkBaseWidget):
    use_ttk = "TFrame"
    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        super().__init__()
        self.layout_weight = 1

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ScrollableFrame(self.tkparent.inner)
        self.ui.set_scroll(self.vertical, self.horizontal)
        super().update(prev)

    def addChild(self, idx, child):
        if idx:
            return
        child_outer = child.outer
        if child_outer:
            child_outer.grid(row=idx, column=0, sticky='nsew')

    def removeChild(self, idx, child):
        child.outer.grid_forget()



from .. import *
from .base import *
from PySide6.QtWidgets import QSizePolicy

class QtScrollArea(QtBaseWidget):
    terminal = False

    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        super().__init__()

    def destroy(self, direct):
        if direct:
            if self.widget:
                self.widget.deleteLater()
                self.widget = None
        super().destroy(direct)

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.widget = prev.widget
        else:
            self.ui = QtWidgets.QScrollArea()
            self.ui.setWidgetResizable(True)
            self.widget = None
            if self.vertical is None:
                self.ui.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
            elif self.vertical:
                self.ui.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
            else:
                self.ui.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            if self.horizontal is None:
                self.ui.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
            elif self.horizontal:
                self.ui.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
            else:
                self.ui.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        super().update(prev)

    def addChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseLayout):
            self.widget = QtWidgets.QWidget()
            self.ui.setWidget(self.widget)
            self.widget.setLayout(child.outer)
            if not hasattr(self.widget, "origResizeEvent"):
                self.widget.origResizeEvent = self.widget.resizeEvent
            self.widget.resizeEvent = self.onUiResized
        elif isinstance(child, QtBaseWidget):
            self.ui.setWidget(child.outer)
            if not hasattr(child.outer, "origResizeEvent"):
                child.outer.origResizeEvent = child.outer.resizeEvent
            child.outer.resizeEvent = self.onUiResized
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
            self.widget.setParent(None)
            self.widget.deleteLater()
            self.widget = None
        elif isinstance(child, QtBaseWidget):
            child.outer.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])

    def onUiResized(self, event):
        node = self.get_node()
        if node.destroyed:
            return
        if node.widget:
            node.widget.origResizeEvent(event)
            node.widget.resizeEvent = self.onUiResized
        else:
            node.children[0].outer.origResizeEvent(event)
            node.children[0].outer.resizeEvent = self.onUiResized
        if node.horizontal is False:
            if isinstance(node.children[0], QtBaseLayout):
                node.outer.setMinimumWidth(node.children[0].outer.sizeHint().width())
            elif isinstance(node.children[0], QtBaseWidget):
                node.outer.setMinimumWidth(node.children[0].outer.sizeHint().width())

        if node.vertical is False:
            if isinstance(node.children[0], QtBaseLayout):
                node.outer.setMinimumHeight(node.children[0].outer.sizeHint().height())
            elif isinstance(node.children[0], QtBaseWidget):
                self.outer.setMinimumHeight(self.children[0].outer.sizeHint().height())



from .. import *
from .base import *

class QtApplication(QPUIView):
    def __init__(self):
        super().__init__()
        self.app = None

    def update(self, prev=None):
        if not self.app:
            from PySide6 import QtWidgets
            self.app = QtWidgets.QApplication([])

        super().update(prev)

    def addChild(self, idx, child):
        child.ui.show()

    def removeChild(self, idx, child):
        child.ui.close()

    def start(self):
        self.app.exec_()



from .. import *
from .base import *

class TScroll(TBase):
    weak_expand_x = True
    weak_expand_y = True
    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        super().__init__()

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = containers.ScrollableContainer()
        v = "auto"
        self.container_y = True
        if self.vertical is True:
            v = "scroll"
        elif self.vertical is False:
            v = "hidden"
            self.container_y = False
            self.weak_hug_y = True

        h = "auto"
        self.container_x = True
        if self.horizontal is True:
            h = "scroll"
        elif self.horizontal is False:
            h = "hidden"
            self.container_x = False
            self.weak_hug_x = True
        self.ui.set_styles(f"overflow-x: {h}; overflow-y: {v};")

        parent = self.tparent
        if parent:
            if parent.strong_expand_x:
                self.strong_expand_x = True
            if parent.strong_expand_y:
                self.strong_expand_y = True

        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, TBase):
            self.inner.mount(child.outer)
            self.t_update_layout()
        else:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, TBase):
            child.tremove()
        else:
            self.removeChild(idx, child.children[0])



from .. import *
from .base import *

class TProgressBar(TBase):
    def __init__(self, progress, maximum=1):
        super().__init__()
        self.progress = progress*100
        self.maximum = maximum*100

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = widgets.ProgressBar(total=self.maximum, show_eta=False)
        self.ui.update(total=self.maximum)
        self.ui.progress = self.progress
        super().update(prev)



from .. import *

from PySide6 import QtCore, QtWidgets, QtGui

class QtViewSignal(QtCore.QObject):
    redraw = QtCore.Signal()

def _apply_params(ui, node):
    styles = {}
    if node.style_fontsize:
        styles["font"] = f"{node.style_fontsize}pt"
    if node.style_color:
        styles["color"] = f"#{node.style_color:06X}"
    if node.style_bgcolor:
        styles["background-color"] = f"#{node.style_bgcolor:06X}"

    HorizontalPolicy = node.qt_params.get("HorizontalPolicy")
    if not HorizontalPolicy is None:
        ui.sizePolicy().setHorizontalPolicy(HorizontalPolicy)

    VerticalPolicy = node.qt_params.get("VerticalPolicy")
    if not VerticalPolicy is None:
        ui.sizePolicy().setVerticalPolicy(VerticalPolicy)

    SizeConstraint = node.qt_params.get("SizeConstraint")
    if not SizeConstraint is None:
        ui.setSizeConstraint(SizeConstraint)

    StyleSheet = node.qt_params.get("StyleSheet", {})
    for k,v in StyleSheet.items():
        styles[k] = v

    if hasattr(ui, "setStyleSheet"):
        ui.setStyleSheet("".join([f"{k}:{v};" for k,v in styles.items()]))

class QPUIView(PUIView):
    def __init__(self):
        super().__init__()
        self.qt_params = {}
        self.signal = QtViewSignal()
        self.signal.redraw.connect(self.update)

    def destroy(self, direct):
        if direct:
            if self.ui: # PUIView doesn't have ui
                self.ui.deleteLater()
        self.ui = None
        super().destroy(direct)

    def redraw(self):
        self.dirty = True
        if self.updating:
            return
        self.updating = True
        self.signal.redraw.emit()

    def update(self, prev=None):
        if self.retired_by:
            return
        self.dirty = False
        super().update(prev)
        _apply_params(self.ui, self)
        self.updating = False
        if self.dirty:
            self.update(prev)

    def qt(self, **kwargs):
        for k,v in kwargs.items():
            self.qt_params[k] = v
        return self

class QtBaseWidget(PUINode):
    terminal = True

    def __init__(self):
        super().__init__()
        self.qt_params = {}

    def destroy(self, direct):
        if direct:
            self.ui.deleteLater()
        self.ui = None
        super().destroy(direct)

    def update(self, prev=None):
        super().update(prev)
        _apply_params(self.ui, self)

    def qt(self, **kwargs):
        for k,v in kwargs.items():
            self.qt_params[k] = v
        return self

class QtBaseLayout(PUINode):
    def __init__(self):
        super().__init__()
        self.qt_params = {}

    def destroy(self, direct):
        if direct:
            self.ui.deleteLater()
        self.ui = None
        super().destroy(direct)

    def update(self, prev=None):
        super().update(prev)
        _apply_params(self.ui, self)

    def addChild(self, idx, child):
        from .layout import QtSpacerItem
        if isinstance(child, QtBaseLayout):
            params = {}
            if not child.layout_weight is None:
                params["stretch"] = child.layout_weight
            self.ui.insertLayout(idx, child.outer, **params)
        elif isinstance(child, QtSpacerItem):
            self.ui.insertItem(idx, child.outer)
        elif isinstance(child, QtBaseWidget):
            params = {}
            if not child.layout_weight is None:
                params["stretch"] = child.layout_weight
            self.ui.insertWidget(idx, child.ui, **params)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        from .layout import QtSpacerItem
        if isinstance(child, QtBaseLayout):
            self.ui.removeItem(child.outer)
        elif isinstance(child, QtSpacerItem):
            self.ui.removeItem(child.outer)
        elif isinstance(child, QtBaseWidget):
            child.ui.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])

    def qt(self, **kwargs):
        for k,v in kwargs.items():
            self.qt_params[k] = v
        return self

class QtBaseFrame(QtBaseWidget):
    terminal = False

    def __init__(self):
        self.widget = None
        super().__init__()

    def destroy(self, direct):
        if direct:
            if self.ui:
                self.ui.deleteLater()
                self.ui = None
            if self.widget:
                self.widget.deleteLater()
                self.widget = None

    def addChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseLayout):
            self.widget = QtWidgets.QWidget()
            self.ui.setWidget(self.widget)
            self.widget.setLayout(child.outer)
        elif isinstance(child, QtBaseWidget):
            self.ui.setWidget(child.outer)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
            self.widget.setParent(None)
        elif isinstance(child, QtBaseWidget):
            child.outer.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])

class QtWrapper(QtBaseWidget):
    def __init__(self, widget, *args):
        super().__init__(*args)
        self.ui = widget

    def destroy(self, direct):
        pass



from .. import *
from .base import *

class TButton(TBase):
    def __init__(self, text, callback=None):
        super().__init__()
        self.text = text
        self.callback = callback

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.label = self.text
        else:
            self.ui = widgets.Button(self.text)
        self.ui.puinode = self
        super().update(prev)


from .. import *
from .base import *

class TVertical(TBase):
    container_y = True
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = containers.Vertical()
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, TBase):
            self.inner.mount(child.outer, before=idx)
        else:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, TBase):
            child.tremove()
        else:
            self.removeChild(idx, child.children[0])

class THorizontal(TBase):
    container_x = True
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = containers.Horizontal()
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, TBase):
            self.inner.mount(child.outer, before=idx)
        else:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, TBase):
            child.tremove()
        else:
            self.removeChild(idx, child.children[0])

class TSpacer(TBase):
    def __init__(self, *args):
        super().__init__(*args)
        self.layout_weight = 1

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = widgets.Static("")
        super().update(prev)


from .. import *
from .base import *
from ..utils import *
from PySide6.QtWidgets import QSizePolicy

class QtCheckbox(QtBaseWidget):
    def __init__(self, text, model):
        super().__init__()
        self.text = text
        self.model = model

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            try:
                self.ui.stateChanged.disconnect()
            except:
                pass
        else:
            self.ui = QtWidgets.QCheckBox()
        self.ui.setText(self.text)
        self.ui.setChecked(bool(self.model.value))
        self.ui.stateChanged.connect(self._stateChanged)
        super().update(prev)

    def _stateChanged(self, value):
        self.model.value = bool(value)



import PySide6
from PySide6.QtWidgets import QSizePolicy, QLayout

from .application import *
from .button import *
from .canvas import *
from .checkbox import *
from .combobox import *
from .label import *
from .layout import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .splitter import *
from .text import *
from .textfield import *
from .window import *
from .mdi import *

Application = QtApplication
Window = QtWindow
HBox = QtHBox
VBox = QtVBox
Label = QtLabel
Button = QtButton
Checkbox = QtCheckbox
RadioButton = QtRadioButton
Canvas = QtCanvas
TextField = QtLineEdit
ProgressBar = QtProgressBar
Scroll = QtScrollArea
Spacer = QtSpacerItem
Text = QtText
Html = QtHtml
MarkDown = QtMarkDown
Combobox = QtComboBox
ComboboxItem = QtComboBoxItem

def QtPUI(func):
    """
    PUI.PySide6.PUI triggers update() by signal/slot
    """
    def func_wrapper(*args):
        class PUIViewWrapper(QPUIView):
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args)

        ret = PUIViewWrapper(func.__name__)
        return ret

    return func_wrapper

PUI = QtPUI

PUI_BACKEND = "PySide6"



from .application import *
from .button import *
from .canvas import *
from .checkbox import *
from .label import *
from .layout import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .text import *
from .textfield import *
from .window import *

class DummyWidget(PUINode):
    supported = False

Application = TkApplication
Window = TkWindow
HBox = TkHBox
VBox = TkVBox
Label = TkLabel
Button = TkButton
Checkbox = TkCheckbutton
RadioButton = TkRadiobutton
Canvas = TkCanvas
Text = TkLabel
TextField = TkEntry
ProgressBar = TkProgressBar
Scroll = TkScroll
Spacer = TkSpacer
Text = TkText
Html = DummyWidget
MarkDown = DummyWidget

PUI_BACKEND = "tkinter"
